/* Copyright (c) 2013-2022. The SimGrid Team. All rights reserved.          */

/* This program is free software; you can redistribute it and/or modify it
 * under the terms of the license (GNU LGPL) which comes with this package. */

#ifndef SMPI_COLLS_PRIVATE_HPP
#define SMPI_COLLS_PRIVATE_HPP

#include "private.hpp"
#include "smpi/mpi.h"
#include "smpi_coll.hpp"
#include "smpi_comm.hpp"
#include "smpi_datatype.hpp"
#include "smpi_op.hpp"
#include "smpi_request.hpp"
#include <cmath>

XBT_LOG_EXTERNAL_DEFAULT_CATEGORY(smpi_colls);

#endif
